# This file is part of ubuntu-bug-triage. See LICENSE file for license info.
"""Bug module."""


class BugTask:
    """Bug Task class."""

    def __init__(self, lp_link):
        """Initialize Bug Task."""
        self._lp = lp_link

        self.src_pkg = self.target.split(' ')[0]

    def __repr__(self):
        """Representation of specific task."""
        return '- %-20s [%s]' % (self.target, self.status)

    @property
    def importance(self):
        """Return importance of bug task."""
        return self._lp.importance

    @property
    def status(self):
        """Return status of bug task."""
        return self._lp.status

    @property
    def target(self):
        """Return target of bug task."""
        return self._lp.bug_target_name


class Bug:
    """Bug class."""

    def __init__(self, lp_link):
        """Initialize bug class."""
        self._lp = lp_link

        self.id = self._lp.id
        self.title = self._lp.title

        self.tasks = []
        for task in self._lp.bug_tasks:
            self.tasks.append(BugTask(task))

    def __eq__(self, other):
        """Two bugs are identical if their ids are equal."""
        return self.id == other.id

    @property
    def affects(self):
        """Create string of all packages affected."""
        return [task.target for task in self.tasks]

    @property
    def url(self):
        """Return URL of bug."""
        return self._lp.web_link

    def to_json(self):
        """Return JSON representation of object."""
        return {
            'id': self.id,
            'title': self.title,
            'affects': self.affects
        }



# This file is part of ubuntu-bug-triage. See LICENSE file for license info.
"""Ubuntu Bug Triage module."""

import argparse
import logging
import sys

from . import UBUNTU_PACKAGE_TEAMS
from .triage import PackageTriage, TeamTriage
from .view import BrowserView, CSVView, JSONView, TerminalView


def parse_args():
    """Set up command-line arguments."""
    parser = argparse.ArgumentParser('ubuntu-bug-triage')
    parser.add_argument(
        'package_or_team', nargs='?', default='ubuntu-server',
        help="""source package name (e.g. cloud-init, lxd) or package team name
        (e.g. ubuntu-openstack, foundations-bugs) to use for search (default:
        ubuntu-server)"""
    )
    parser.add_argument(
        'days', nargs='?', type=int, default=1,
        help="""days of updated bugs to triage"""
    )
    parser.add_argument(
        '--anon', action='store_true',
        help='Anonymous login to Launchpad'
    )
    parser.add_argument(
        '--csv', action='store_true',
        help='output as CSV'
    )
    parser.add_argument(
        '--debug', action='store_true',
        help='additional logging output'
    )
    parser.add_argument(
        '--json', action='store_true',
        help='output as JSON'
    )
    parser.add_argument(
        '--open', action='store_true',
        help='open resulting bugs in web browser'
    )
    parser.add_argument(
        '--include-project', '-p', action='store_true',
        help='include project bugs in output'
    )

    return parser.parse_args()


def setup_logging(debug):
    """Set up logging."""
    logging.basicConfig(
        stream=sys.stdout,
        format='%(message)s',
        level=logging.DEBUG if debug else logging.INFO
    )


def launch():
    """Launch ubuntu-bug-triage."""
    args = parse_args()
    setup_logging(args.debug)

    if args.package_or_team in UBUNTU_PACKAGE_TEAMS:
        if args.include_project:
            logging.getLogger(__name__).warning(
                "N.B. --include-project has no effect when running against a"
                " package team")
        triage = TeamTriage(args.package_or_team, args.days, args.anon)
    else:
        triage = PackageTriage(args.package_or_team, args.days, args.anon,
                               args.include_project)

    bugs = triage.updated_bugs()
    if args.csv:
        CSVView(bugs)
    elif args.json:
        JSONView(bugs)
    else:
        TerminalView(bugs)

    if args.open:
        BrowserView(bugs)


if __name__ == '__main__':
    sys.exit(launch())



# This file is part of ubuntu-bug-triage. See LICENSE file for license info.
"""Test bug module."""
import random

from .bug import Bug, BugTask


class MockBugTaskRandom:
    """Create a task with random valid options."""

    def __init__(self):
        """Initialize class."""
        self.importance = random.choice(
            ['Wishlist', 'Low', 'Medium', 'High', 'Critical', 'Undecided']
        )
        self.status = random.choice(
            ['New', 'Incomplete', 'Opinion', 'Invalid', 'Won\'t Fix',
             'Confirmed', 'Triaged', 'In Progress', 'Fix Committed',
             'Fix Released']
        )
        self.bug_target_name = random.choice(
            ['Trusty', 'Xenial', 'Bionic', 'Disco']
        )


class MockBugTask:
    """Create mock bug task with specific items."""

    def __init__(self):
        """Initialize class."""
        self.importance = 'Wishlist'
        self.status = 'Triaged'
        self.bug_target_name = 'Bionic'


class MockBug:
    """Create fake bug with random tasks."""

    def __init__(self):
        """Initialize class."""
        self.id = '1761240'
        self.title = 'SRU pollinate 4.33'
        self.status = 'Fix Released'
        self.importance = 'Low'
        self.bug_target_name = 'pollinate (Ubuntu)'
        self.web_link = (
            'https://bugs.launchpad.net/ubuntu/+source/pollinate/+bug/1761240'
        )
        num_tasks = range(random.randint(1, 8))
        self.bug_tasks = [MockBugTaskRandom() for _ in num_tasks]


class MockBugNoTasks:
    """Create bug with no tasks."""

    def __init__(self):
        """Initialize class."""
        self.id = '1761240'
        self.title = 'SRU pollinate 4.33'
        self.status = 'Fix Released'
        self.importance = 'Low'
        self.bug_target_name = 'pollinate (Ubuntu)'
        self.web_link = (
            'https://bugs.launchpad.net/ubuntu/+source/pollinate/+bug/1761240'
        )
        self.bug_tasks = []


class MockBug1234567:
    """Fake bug with specific data."""

    def __init__(self):
        """Initialize class."""
        self.id = '1234567'
        self.title = 'Test Bug'
        self.status = 'In Progress'
        self.importance = 'Critical'
        self.bug_target_name = 'cloud-init (Ubuntu)'
        self.web_link = (
            'https://bugs.launchpad.net/ubuntu/+source/cloud-init/+bug/1234567'
        )
        self.bug_tasks = []


class MockBug1111111:
    """Another fake bug with specific data."""

    def __init__(self):
        """Initialize class."""
        self.id = '1111111'
        self.title = 'Test Bug 2'
        self.status = 'Triaged'
        self.importance = 'Wishlist'
        self.bug_target_name = 'curtin (Ubuntu)'
        self.web_link = (
            'https://bugs.launchpad.net/ubuntu/+source/cloud-init/+bug/1111111'
        )
        self.bug_tasks = []


def test_bug_class():
    """Test Bug Class."""
    assert Bug(MockBug())


def test_bug_class_no_tasks():
    """Test Bug Class with no tasks."""
    assert Bug(MockBugNoTasks())


def test_bug_same():
    """Verify bug compare works."""
    assert Bug(MockBug1234567()) == Bug(MockBug1234567())


def test_bug_not_same():
    """Verify bug compare works negative case."""
    assert Bug(MockBug1234567()) != Bug(MockBug1111111())


def test_affects():
    """Verify affects is an array."""
    bug = Bug(MockBug())
    assert isinstance(bug.affects, type([]))


def test_no_affects():
    """Verify no affects."""
    bug = Bug(MockBugNoTasks())
    assert isinstance(bug.affects, type([]))
    assert bug.affects == []


def test_url():
    """Verify url is a string."""
    bug = Bug(MockBug())
    assert isinstance(bug.url, str)


def test_json():
    """Verify JSON output."""
    bug = Bug(MockBug1234567())
    json_output = {
        'id': bug.id,
        'title': bug.title,
        'affects': bug.affects
    }
    assert bug.to_json() == json_output


def test_bug_task():
    """Verify bug task attributes."""
    task = BugTask(MockBugTask())
    assert task.importance == 'Wishlist'
    assert task.status == 'Triaged'
    assert repr(task) == '- %-20s [%s]' % (task.target, task.status)



# This file is part of ubuntu-bug-triage. See LICENSE file for license info.
"""Ubuntu Bug Triage module."""

# This is a hard-coded list of teams that produce a lot of bugs and may
# not want to be included in the full team triage.
#
BLACKLIST = {
    'ubuntu-server': [
        'cloud-init',
        'curtin',
        'juju',
        'juju-core',
        'lxc',
        'lxd',
        'maas'
    ]
}

# This is hard-coded as it took way too long to download the json file
# from the following URL:
#
# https://people.canonical.com/~ubuntu-archive/package-team-mapping.json
#
UBUNTU_PACKAGE_TEAMS = [
    'pkg-ime',
    'ubuntu-openstack',
    'documentation-packages',
    'kubuntu-bugs',
    'checkbox-bugs',
    'foundations-bugs',
    'maas-maintainers',
    'ubuntu-server',
    'ubuntu-printing',
    'ubuntu-security-bugs',
    'mir-team',
    'kernel-packages',
    'unsubscribed',
    'desktop-packages',
    'translators-packages',
    'snappy-dev',
]



# This file is part of ubuntu-bug-triage. See LICENSE file for license info.
"""Triage module."""

import json
import logging
import os
import textwrap
import time
import webbrowser

from tabulate import tabulate


class BaseView:
    """Base view class."""

    def __init__(self):
        """Initialize base view."""
        self._log = logging.getLogger(__name__)


class BrowserView(BaseView):
    """Browser view class.

    This class uses the webbrowser from the Python standard library.
    There is no way to know if a page has fully loaded or know when
    a browser is even open. As such this is fire and forget behavior
    and it leads to errors on opening pages.

    I looked into selenium, but found mixed results and had to figure
    out what browser to use in the first place and went with the crude
    sleep statements below.
    """

    def __init__(self, bugs):
        """Initialize browser view."""
        super().__init__()

        self._log.debug('opening bugs in browser')
        self.opened = False
        for bug in bugs:
            self._open_url(bug.url)

    def _open_url(self, url):
        """Open a given url."""
        if self.opened:
            webbrowser.open_new_tab(url)
            time.sleep(1.2)
        else:
            webbrowser.open(url)
            self.opened = True
            time.sleep(5)


class CSVView(BaseView):
    """CSV view class."""

    def __init__(self, bugs):
        """Initialize CSV view."""
        super().__init__()

        print('id,affects,title')
        for bug in bugs:
            print('LP: #%s,"%s",%s' % (
                bug.id, ','.join(bug.affects), bug.title
            ))


class JSONView(BaseView):
    """JSON view class."""

    def __init__(self, bugs):
        """Initialize JSON view."""
        super().__init__()

        print(json.dumps(
            [bug.to_json() for bug in bugs],
            indent=4,
            ensure_ascii=False
        ))


class TerminalView(BaseView):
    """Terminal view class."""

    def __init__(self, bugs):
        """Initialize terminal view."""
        super().__init__()

        _, self.term_columns = os.popen('stty size', 'r').read().split()

        # The table has a set number characters that always exist:
        #     4 for '|' boarders
        #     6 for spaces around content
        #    12 for 'LP: #1234567'
        width_table = 22

        # This seemed liked the best middle of the road value
        width_affects = 20

        # The title width will expand as the terminal expands,
        # but will expect a width no less than 80
        width_title = (max(80, int(self.term_columns)) -
                       width_affects - width_table)

        table = []
        for bug in bugs:
            affects = [
                textwrap.fill(
                    pkg.replace('(Ubuntu)', '(U)').replace('(Debian)', '(D)'),
                    width=width_affects,
                    subsequent_indent='    '
                ) for pkg in bug.affects
            ]

            table.append([
                'LP: #%s' % bug.id,
                '\n'.join(affects),
                textwrap.fill(bug.title, width=width_title, max_lines=4),
            ])

        print(tabulate(table, ['id', 'affects', 'title'], 'grid'))
